/* Copyright (C) 1992, 1995, 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* Adapted to MiNTLib by Guido Flohr <gufl0000@stud.uni-sb.de>.  */

#ifndef _IEEE754_H

#define _IEEE754_H 1

#include <features.h>

#include <endian.h>

__BEGIN_DECLS

#ifdef __MSHORT__
# define __uint32 unsigned long int
#else
# define __uint32 unsigned int
#endif

union ieee754_float
  {
    float f;

    /* This is the IEEE 754 single-precision format.  */
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	__uint32 negative:1;
	__uint32 exponent:8;
	__uint32 mantissa:23;
#endif				/* Big endian.  */
#if	__BYTE_ORDER == __LITTLE_ENDIAN
	__uint32 mantissa:23;
	__uint32 exponent:8;
	__uint32 negative:1;
#endif				/* Little endian.  */
      } ieee;

    /* This format makes it easier to see if a NaN is a signalling NaN.  */
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	__uint32 negative:1;
	__uint32 exponent:8;
	__uint32 quiet_nan:1;
	__uint32 mantissa:22;
#endif				/* Big endian.  */
#if	__BYTE_ORDER == __LITTLE_ENDIAN
	__uint32 mantissa:22;
	__uint32 quiet_nan:1;
	__uint32 exponent:8;
	__uint32 negative:1;
#endif				/* Little endian.  */
      } ieee_nan;
  };

#define IEEE754_FLOAT_BIAS	0x7f /* Added to exponent.  */


union ieee754_double
  {
    double d;

    /* This is the IEEE 754 double-precision format.  */
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	__uint32 negative:1;
	__uint32 exponent:11;
	/* Together these comprise the mantissa.  */
	__uint32 mantissa0:20;
	__uint32 mantissa1:32;
#endif				/* Big endian.  */
#if	__BYTE_ORDER == __LITTLE_ENDIAN
	/* Together these comprise the mantissa.  */
	__uint32 mantissa1:32;
	__uint32 mantissa0:20;
	__uint32 exponent:11;
	__uint32 negative:1;
#endif				/* Little endian.  */
      } ieee;

    /* This format makes it easier to see if a NaN is a signalling NaN.  */
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	__uint32 negative:1;
	__uint32 exponent:11;
	__uint32 quiet_nan:1;
	/* Together these comprise the mantissa.  */
	__uint32 mantissa0:19;
	__uint32 mantissa1:32;
#else
	/* Together these comprise the mantissa.  */
	__uint32 mantissa1:32;
	__uint32 mantissa0:19;
	__uint32 quiet_nan:1;
	__uint32 exponent:11;
	__uint32 negative:1;
#endif
      } ieee_nan;
  };

#define IEEE754_DOUBLE_BIAS	0x3ff /* Added to exponent.  */


union ieee854_long_double
  {
    long double d;

    /* This is the IEEE 854 double-extended-precision format.  */
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	__uint32 negative:1;
	__uint32 exponent:15;
	__uint32 empty:16;
	__uint32 mantissa0:32;
	__uint32 mantissa1:32;
#endif
#if	__BYTE_ORDER == __LITTLE_ENDIAN
	__uint32 mantissa1:32;
	__uint32 mantissa0:32;
	__uint32 exponent:15;
	__uint32 negative:1;
	__uint32 empty:16;
#endif
      } ieee;

    /* This is for NaNs in the IEEE 854 double-extended-precision format.  */
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	__uint32 negative:1;
	__uint32 exponent:15;
	__uint32 empty:16;
	__uint32 one:1;
	__uint32 quiet_nan:1;
	__uint32 mantissa0:30;
	__uint32 mantissa1:32;
#endif
#if	__BYTE_ORDER == __LITTLE_ENDIAN
	__uint32 mantissa1:32;
	__uint32 mantissa0:30;
	__uint32 quiet_nan:1;
	__uint32 one:1;
	__uint32 exponent:15;
	__uint32 negative:1;
	__uint32 empty:16;
#endif
      } ieee_nan;
  };

#define IEEE854_LONG_DOUBLE_BIAS 0x3fff

#undef __uint32

__END_DECLS

#endif /* ieee754.h */
