/*  pwd.h -- MiNTLib.
    Copyright (C) 1999 Guido Flohr <gufl0000@stud.uni-sb.de>

    This file is part of the MiNTLib project, and may only be used
    modified and distributed under the terms of the MiNTLib project
    license, COPYMINT.  By continuing to use, modify, or distribute
    this file you indicate that you have read the license and
    understand and accept it fully.
*/

#ifndef _PWD_H
#define _PWD_H
/* File to describe password file fields.  */

#ifndef _COMPILER_H
# include <compiler.h>
#endif

#if defined(__USE_SVID) || defined(__USE_GNU)
# ifndef _STDIO_H
# define __need_FILE
#  include <stdio.h>
# endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __USE_XOPEN
/* The Single Unix specification says that some more types are
   available here.  */
# ifndef gid_t
typedef _GID_T gid_t;
#  define gid_t gid_t
# endif

# ifndef uid_t
typedef _UID_T uid_t;
#  define uid_t uid_t
# endif
#endif

struct passwd
{
  char    *pw_name;    /* Username.  */
  char    *pw_passwd;  /* Encrypted password.  */
  _UID_T  pw_uid;      /* Numerical user ID.  */
  _GID_T  pw_gid;      /* Numerical primary group ID.  */
  char    *pw_age;     /* Password aging information if available.  */
  char    *pw_comment; /* Duplicate of pw_gecos.  */
  char    *pw_gecos;   /* Real name resp. gecos (comment) field.  */
  char    *pw_dir;     /* Home directory.  */
  char    *pw_shell;   /* Login shell.  */
};

__EXTERN struct passwd	*getpwnam __PROTO((const char *));
__EXTERN struct passwd	*getpwuid __PROTO((_UID_T));

#if defined(__USE_SVID) || defined(__USE_MISC)
__EXTERN int		getpw		__PROTO((_UID_T, char *));
#endif

#if defined (__USE_SVID) || defined (__USE_MISC) || defined (__USE_XOPEN_EXTENDED)
__EXTERN void		setpwent	__PROTO((void));
__EXTERN void		endpwent	__PROTO((void));
__EXTERN struct passwd	*getpwent	__PROTO((void));
#endif

#ifdef __USE_BSD
/* FIXME:  This is just a guess. Where does setpwfile belong?  */
__EXTERN void		setpwfile	__PROTO((char *));
#endif

#ifdef __USE_SVID
__EXTERN struct passwd	*fgetpwent	__PROTO((FILE *));
__EXTERN int		putpwent __PROTO((const struct passwd *, FILE *));
#endif

# if defined __USE_SVID || defined __USE_MISC || defined __USE_XOPEN_EXTENDED
__EXTERN int getpwent_r __PROTO ((struct passwd* __resultbuf, char* __buffer, 
		                _SIZE_T __buflen, struct passwd** __result));
# endif

#if defined __USE_POSIX || defined __USE_MISC
__EXTERN int getpwuid_r __PROTO ((_UID_T __uid, struct passwd* __resultbuf, 
                                  char* __buffer, _SIZE_T __buflen, 
                                  struct passwd** __result));

__EXTERN int getpwnam_r __PROTO ((const char* __name, struct passwd* __resultbuf,
			          char* __buffer, _SIZE_T __buflen,
			          struct passwd** __result));
#endif


# ifdef	__USE_SVID
/* Read an entry from STREAM.  This function is not standardized and
   probably never will.  */
__EXTERN int fgetpwent_r __PROTO ((FILE* __stream, struct passwd* __resultbuf,
			     char* __buffer, _SIZE_T __buflen,
			     struct passwd** __result));
# endif

#ifdef __cplusplus
}
#endif

#endif /* _PWD_H */
