/* from dLibs, by Dale Schumacher */

/*
 *  CTYPE.C	Character classification and conversion
 */

#define __NO_CTYPE

#include <ctype.h>
#include <limits.h>

#undef	toupper /* note that in gcc we have a safe version of these, */
#undef  tolower	/* but its better to leave these as routines in case */
		/* some code uses these as function pointers --  i   */
		/* have seen code that does.			     */

static unsigned char _myctype[UCHAR_MAX+1] =
	{
	_CTc, _CTc, _CTc, _CTc,				/* 0x00..0x03 */
	_CTc, _CTc, _CTc, _CTc,				/* 0x04..0x07 */
	_CTc, _CTc|_CTs|_CTb, _CTc|_CTs, _CTc|_CTs,	/* 0x08..0x0B */
	_CTc|_CTs, _CTc|_CTs, _CTc, _CTc,		/* 0x0C..0x0F */

	_CTc, _CTc, _CTc, _CTc,				/* 0x10..0x13 */
	_CTc, _CTc, _CTc, _CTc,				/* 0x14..0x17 */
	_CTc, _CTc, _CTc, _CTc,				/* 0x18..0x1B */
	_CTc, _CTc, _CTc, _CTc,				/* 0x1C..0x1F */

	_CTs|_CTb, _CTp, _CTp, _CTp,			/* 0x20..0x23 */
	_CTp, _CTp, _CTp, _CTp,				/* 0x24..0x27 */
	_CTp, _CTp, _CTp, _CTp,				/* 0x28..0x2B */
	_CTp, _CTp, _CTp, _CTp,				/* 0x2C..0x2F */

	_CTd|_CTx, _CTd|_CTx, _CTd|_CTx, _CTd|_CTx,	/* 0x30..0x33 */
	_CTd|_CTx, _CTd|_CTx, _CTd|_CTx, _CTd|_CTx,	/* 0x34..0x37 */
	_CTd|_CTx, _CTd|_CTx, _CTp, _CTp,		/* 0x38..0x3B */
	_CTp, _CTp, _CTp, _CTp,				/* 0x3C..0x3F */

	_CTp, _CTu|_CTx, _CTu|_CTx, _CTu|_CTx,		/* 0x40..0x43 */
	_CTu|_CTx, _CTu|_CTx, _CTu|_CTx, _CTu,		/* 0x44..0x47 */
	_CTu, _CTu, _CTu, _CTu,				/* 0x48..0x4B */
	_CTu, _CTu, _CTu, _CTu,				/* 0x4C..0x4F */

	_CTu, _CTu, _CTu, _CTu,				/* 0x50..0x53 */
	_CTu, _CTu, _CTu, _CTu,				/* 0x54..0x57 */
	_CTu, _CTu, _CTu, _CTp,				/* 0x58..0x5B */
	_CTp, _CTp, _CTp, _CTp,				/* 0x5C..0x5F */

	_CTp, _CTl|_CTx, _CTl|_CTx, _CTl|_CTx,		/* 0x60..0x63 */
	_CTl|_CTx, _CTl|_CTx, _CTl|_CTx, _CTl,		/* 0x64..0x67 */
	_CTl, _CTl, _CTl, _CTl,				/* 0x68..0x6B */
	_CTl, _CTl, _CTl, _CTl,				/* 0x6C..0x6F */

	_CTl, _CTl, _CTl, _CTl,				/* 0x70..0x73 */
	_CTl, _CTl, _CTl, _CTl,				/* 0x74..0x77 */
	_CTl, _CTl, _CTl, _CTp,				/* 0x78..0x7B */
	_CTp, _CTp, _CTp, _CTc,				/* 0x7C..0x7F */

	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x80..0x8F */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x90..0x9F */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xA0..0xAF */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xB0..0xBF */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xC0..0xCF */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xD0..0xDF */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xE0..0xEF */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0  /* 0xF0..0xFF */
	};

unsigned char *_ctype = &(_myctype[0]);

int isalnum (c)
     register int c;
{
 return (int) (_ctype[(unsigned char)(c)]&(_CTu|_CTl|_CTd));
}

int 	
isalpha (c)
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&(_CTu|_CTl));
}

int 	
isascii (c)	
     register int c;
{	
  return (int) !((c)&~0x7F);
}

int 	
toascii(c)	
     register int c;
{	
  return (int) ((c)&0x7F);
}

int 	
iscntrl(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTc);
}

int 	
isdigit(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTd);
}

int 	
isgraph(c)	
     register int c;
{	
  return (int) (!(_ctype[(unsigned char)(c)]&(_CTc|_CTs)) && (_ctype[(unsigned char)(c)]));
}

int 	
islower(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTl);
}

int  
isprint(c)      
     register int c;
{	
  return (int) (!(_ctype[(unsigned char)(c)]&_CTc) && (_ctype[(unsigned char)(c)]));
}

int 	
ispunct(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTp);
}

int 	
isspace(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTs);
}

int 	
isupper(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTu);
}

int 	
isxdigit(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTx);
}

int 
isblank(c)	
     register int c;
{	
  return (int) (_ctype[(unsigned char)(c)]&_CTb);
}
