/*  wait4.c -- MiNTLib.
    Copyright (C) 1999 Guido Flohr <gufl0000@stud.uni-sb.de>

    This file is part of the MiNTLib project, and may only be used
    modified and distributed under the terms of the MiNTLib project
    license, COPYMINT.  By continuing to use, modify, or distribute
    this file you indicate that you have read the license and
    understand and accept it fully.
*/

#include <signal.h>
#include <errno.h>
#include <string.h>
#include <mintbind.h>

#ifdef __TURBOC__
# include <sys\wait.h>
# include <sys\types.h>
# include <sys\resource.h>
#else
# include <sys/wait.h>
# include <sys/types.h>
# include <sys/resource.h>
#endif

static int have_Pwaitpid = 1;

extern long __waitval, __waittime;

pid_t __wait4 (pid, stat_loc, options, usage)
  pid_t pid;
  __WP stat_loc;
  int options;
  struct rusage* usage;
{
  long retval;
  long lusage[8];  /* For current MiNT versions 2 longs should 
                    * be enough.  The MiNTLib has 8 longs.
                    * Don't know why but better leave it as
                    * it is.
                    */
  pid_t child_pid;
  
  if (have_Pwaitpid) {
    retval = Pwaitpid (pid, options, lusage);
    if (retval == -ENOSYS)
      have_Pwaitpid = 0;
  }
  
  if (!have_Pwaitpid) {
    retval = __waitval;
    __waitval = -ESRCH;
    lusage[0] = __waittime;
    lusage[1] = 0;
  }

  /* That's all we can get from the OS.  */
  if (retval < 0) {
    if ((options & WNOHANG) && retval == -ENOENT)
      return 0;  /* For MiNT this is an error.  For us it isn't.  */

    errno = -retval;
    return -1; 
  }

  child_pid = (pid_t) (retval >> 16);

  /* Mask out the child pid.  */
  retval &= 0xffff;
  
  if (stat_loc != NULL) {
    if (((char) (retval & 0xff)) == (char) -32 && Pdomain (0) != 1) {
      /* Special kludge in the MiNT kernel.  */
      *stat_loc = SIGINT;
    } else if (((retval & 0x7f) == 0x7f) && ((retval & 0xff00) != 0)) {
      *stat_loc = retval;
    } else {
      /* MiNT returns the lower 2 bytes swapped.  */
      *stat_loc = ((retval & 0xff00) >> 8) | ((retval & 0xff) << 8);
    }
  }
  
  if (usage) {
    memset (usage, 0, sizeof (struct rusage));
    usage->ru_utime.tv_sec = lusage[0] / 1000;
    usage->ru_utime.tv_usec = 1000 * (lusage[0] % 1000);
    usage->ru_stime.tv_sec = lusage[1] / 1000;
    usage->ru_stime.tv_usec = 1000 * (lusage[1] % 1000);
  }
  
  return child_pid;
}

#ifdef HAVE_WEAK_SYMBOLS
weak_alias (__wait4, wait4)
#endif
